from typing import Callable

import libtmux

from lenses import lens

from amino import List, _, __, Just, Maybe, Map, Right, Empty, Either, Boolean
from amino.lazy import lazy
from amino.task import Task
from amino.anon import L

from ribosome.machine import may_handle, handle, Quit
from ribosome.record import field, list_field, Record
from ribosome.machine.base import DataEitherTask, UnitTask

from myo.state import MyoComponent, MyoTransitions
from myo.plugins.tmux.message import (TmuxOpenPane, TmuxRunCommand,
                                      TmuxCreatePane, TmuxCreateLayout,
                                      TmuxCreateSession, TmuxSpawnSession,
                                      TmuxFindVim, TmuxInfo, TmuxLoadDefaults,
                                      TmuxClosePane, TmuxRunShell,
                                      TmuxRunLineInShell, StartWatcher,
                                      WatchCommand, QuitWatcher, SetCommandLog,
                                      TmuxPack)
from myo.plugins.core.main import StageI
from myo.ui.tmux.pane import Pane, VimPane
from myo.ui.tmux.layout import LayoutDirections, Layout, VimLayout
from myo.ui.tmux.session import Session
from myo.ui.tmux.server import Server
from myo.util import parse_int, view_params
from myo.ui.tmux.window import VimWindow, Window
from myo.ui.tmux.facade import LayoutFacade, PaneFacade
from myo.ui.tmux.view import View
from myo.plugins.core.message import AddDispatcher
from myo.plugins.tmux.dispatch import TmuxDispatcher
from myo.ui.tmux.util import format_state, Ident
from myo.ui.tmux.pane_path import PanePathMod, ppm_id
from myo.plugins.command.message import SetShellTarget
from myo.command import Command, Shell
from myo.plugins.tmux.watcher import Watcher, Terminated

_is_vim_window = lambda a: isinstance(a, VimWindow)
_is_vim_layout = lambda a: isinstance(a, VimLayout)
_is_vim_pane = lambda a: isinstance(a, VimPane)


def _ident_ppm(ident: Ident):
    return PanePathMod(pred=__.has_ident(ident))


class TmuxState(Record):
    server = field(Server)
    sessions = list_field()
    windows = list_field()
    instance_id = field(str, initial='', factory=L(List.random_string)(5))

    @property
    def vim_window(self):
        return self.windows.find(_is_vim_window)

    @property
    def vim_window_lens(self):
        return (self.windows.find_lens_pred(_is_vim_window) /
                lens(self).windows.add_lens)

    @property
    def vim_layout(self):
        return self.vim_window // __.root.layouts.find(_is_vim_layout)

    @property
    def vim_layout_lens(self):
        return (
            self.vim_layout //
            __.layouts.find_lens_pred(_is_vim_layout) /
            lens(self).windows.add_lens
        )

    @property
    def vim_pane(self):
        return self.vim_layout / _.panes / _.head

    @property
    def all_panes(self):
        return self.windows // _.root.all_panes

    def pane(self, ident):
        return self.all_panes.find(__.has_ident(ident))

    @property
    def possibly_open_panes(self):
        return (
            self.all_panes
            .filter(_.id.is_just)
            .filter(lambda a: not _is_vim_pane(a))
        )


class TmuxTransitions(MyoTransitions):

    def _state(self, data):
        return data.sub_state(self.name, self.machine.new_state)

    @property
    def state(self):
        return self._state(self.data)

    def _with_sub(self, data, state):
        return data.with_sub_state(self.name, state)

    def with_sub(self, state):
        return self._with_sub(self.data, state)

    def _with_root(self, data, state, root):
        l = state.vim_window_lens / _.root
        new_state = l / __.set(root) | state
        return self._with_sub(data, new_state)

    def _with_vim_window(self, data, state, win):
        new_state = state.vim_window_lens / __.set(win) | state
        return self._with_sub(data, new_state)

    def with_sub_and_msgs(self, state, msgs):
        return (self.with_sub(state),) + msgs

    @property
    def server(self):
        return self.state.server

    @property
    def layouts(self):
        return self.machine.layouts

    @property
    def panes(self):
        return self.machine.panes

    @may_handle(StageI)
    def stage_1(self):
        ''' Initialize the state. If it doesn't exist, Env will create
        it using the constructor function supplied in *Plugin.state*
        '''
        default = self.pflags.tmux_use_defaults.maybe(TmuxLoadDefaults())
        msgs = List(AddDispatcher(), self.with_sub(self.state), TmuxFindVim())
        return msgs + default.to_list + List(StartWatcher())

    @may_handle(AddDispatcher)
    def add_dispatcher(self):
        return self.data + TmuxDispatcher()

    @may_handle(TmuxFindVim)
    def find_vim(self):
        vim_pid = self._vim_pid
        vim_pane = vim_pid // self._find_vim_pane
        id = (
            (self.vim.pvar('tmux_force_vim_pane_id') // parse_int)
            .or_else(vim_pane // _.id_i)
        )
        wid = (
            self.vim.pvar('tmux_force_vim_win_id')
            .or_else(vim_pane // _.window_id_i)
            .to_either('no vim win id')
        )
        sid = (
            self.vim.pvar('tmux_force_vim_session_id')
            .or_else(vim_pane // _.session_id_i)
            .to_either('no vim session id')
        )
        vim_w = self.vim.pvar('tmux_vim_width').or_else(Just(85))
        pane = VimPane(id=id.to_maybe, name='vim', pid=vim_pid.to_maybe,
                       window_id=wid, session_id=sid)
        vim_layout = VimLayout(name='vim',
                               direction=LayoutDirections.vertical,
                               panes=List(pane), fixed_size=vim_w)
        root = Layout(name='root',
                      direction=LayoutDirections.horizontal,
                      layouts=List(vim_layout))
        win = VimWindow(id=wid.to_maybe, root=root)
        new_state = self.state.append1.windows(win)
        return self.with_sub(new_state)

    @may_handle(TmuxLoadDefaults)
    def load_defaults(self):
        main = TmuxCreateLayout('main', options=Map(parent='root'))
        make = TmuxCreatePane('make', options=Map(parent='main'))
        return main, make

    @may_handle(StartWatcher)
    def start_watcher(self):
        return UnitTask(Task(self.machine.watcher.start))

    @handle(TmuxCreateSession)
    def create_session(self):
        return (
            self.msg.options.get('name') /
            (lambda n: Session(id=n)) /
            List /
            self.state.append.sessions /
            self.with_sub
        )

    @may_handle(TmuxSpawnSession)
    def spawn_session(self):
        s = self.state.session(self.msg.id)
        self.server.new_session(session_name=s.x)

    @handle(TmuxCreateLayout)
    def create_layout(self):
        opts = self.msg.options
        dir_s = opts.get('direction') | 'vertical'
        direction = LayoutDirections.parse(dir_s)
        layout = Layout(name=self.msg.name, direction=direction,
                        **view_params(opts))
        return self._add_to_layout(opts.get('parent'), _.layouts, layout)

    @handle(TmuxCreatePane)
    def create_pane(self):
        opts = self.msg.options
        pane = Pane(name=self.msg.name, **view_params(opts))
        return self._add_to_layout(opts.get('parent'), _.panes, pane)

    @may_handle(TmuxOpenPane)
    def open(self):
        return self._open_pane_ppm(self.msg.name)

    @may_handle(TmuxClosePane)
    def close(self):
        return self._close_pane_ppm(self.msg.name)

    @may_handle(TmuxRunCommand)
    def run_command(self):
        command = self.msg.command
        shell = command.shell // self.data.shell
        options = self.msg.options
        return shell.cata(
            L(self._run_in_shell)(command, _, options),
            L(self._run_command)(command, options)
        )

    @may_handle(TmuxRunShell)
    def run_shell(self):
        shell = self.msg.shell
        default = Map(shell.target.map(lambda a: ('pane', a)).to_list)
        opts = default ** self.msg.options
        return TmuxRunCommand(command=shell, options=opts)

    @handle(SetShellTarget)
    def set_shell_target(self):
        target = self._pane(self.msg.target) / _.uuid
        setter = __.modify(__.set(target=target))
        return (self.data.command_lens(self.msg.shell.uuid)).map(setter)

    @handle(TmuxRunLineInShell)
    def run_line_in_shell(self):
        opt = self.msg.options
        return (
            opt.get('line') /
            (lambda a: Command(name='temp', line=a)) /
            L(self._run_in_shell)(_, self.msg.shell, opt)
        )

    @may_handle(PanePathMod)
    def pane_path_mod(self):
        return self._run_ppm(self.msg)

    @may_handle(Quit)
    def quit(self):
        t = self.state.possibly_open_panes // _.id / self.panes.close_id
        return QuitWatcher(), UnitTask(t.sequence(Task))

    @may_handle(QuitWatcher)
    def quit_watcher(self):
        return UnitTask(Task(self.machine.watcher.stop))

    @may_handle(TmuxInfo)
    def info(self):
        self.log.info('\n'.join(format_state(self.state)))

    @may_handle(Terminated)
    def terminated(self):
        return self._pane_mod(self.msg.pane.uuid, __.setter.pid(Empty()))

    @may_handle(SetCommandLog)
    def set_command_log(self):
        def set_log(data):
            log_path = self._pane(self.msg.pane_ident) // _.log_path
            cmd = data.command_lens(self.msg.cmd_ident)
            res = cmd / (lambda c: c.modify(__.set(log_path=log_path)))
            return res.to_either('command not found')
        # NOTE Must run this as a task so it is executed after the cmd runner
        return DataEitherTask(_ / set_log)

    @may_handle(TmuxPack)
    def pack(self):
        return self._window_task(self.layouts.pack_window)

    def _window_task(self, f: Callable[[Window], Task[Either[str, Window]]]):
        return DataEitherTask(_ // L(self._wrap_window)(_, f))

    def _wrap_window(self, data, callback):
        state = self._state(data)
        win = Task.from_maybe(state.vim_window, 'no vim window')
        update = __.map(L(self._with_vim_window)(data, state, _))
        return win // callback / update

    def _run_ppm(self, ppm):
        return self._window_task(ppm.run)

    def _open_pane_ppm(self, name: Ident):
        # cannot reference self.layouts.pack_path directly, because panes
        # are cached
        return _ident_ppm(name) / self._open_pane / self._pack_path

    def _close_pane_ppm(self, name: Ident):
        return _ident_ppm(name) / self._close_pane / self._pack_path

    def _run_command_ppm(self, command, opt: Map):
        in_shell = Boolean('shell' in opt)
        pane_ident = opt.get('pane') | self._default_pane_name
        def check_running(path):
            return Task.now(Right(path)) if in_shell else (
                self.panes.ensure_not_running(path.pane) /
                __.replace(path)
            )
        def pipe(path):
            return (
                self.panes.pipe(path.pane, self.state.instance_id) /
                path.setter.pane /
                Right
            )
        def run(path):
            return (
                self._run_command_line(command.line, path.pane, opt) /
                (lambda a: Right(path))
            )
        def pid(path):
            return (
                self.panes.command_pid(path.pane) /
                (lambda a: path.map_pane(__.set(pid=a))) /
                Right
            )
        def watch(path):
            msg = WatchCommand(command, path.pane)
            return (
                Task.call(self.machine.watcher.send, msg)
                .replace(Right(path))
            )
        runner = (
            (self._open_pane_ppm(pane_ident) + check_running) /
            in_shell.no.maybe(pipe).get_or_else(lambda: ppm_id) /
            run /
            pid /
            watch
        )
        set_log = command.transient.no.maybe(
            SetCommandLog(command.uuid, pane_ident))
        return set_log.to_list.cons(runner)

    def _open_pane(self, w):
        return self.layouts.open_pane(w)

    def _close_pane(self, w):
        return self.layouts.close_pane_path(w)

    def _pack_path(self, path):
        return self.layouts.pack_path(path)

    def _layout_lens_bound(self, name):
        def sub(a):
            return a.layouts.find_lens(ll) / lens().layouts.add_lens
        def ll(l):
            return Just(lens()) if l.name == name else sub(l)
        return (
            self.state.windows
            .find_lens(lambda a: ll(a.root) / lens().root.add_lens) /
            lens(self.state).windows.add_lens
        )

    def _add_to_layout(self, parent: Maybe[str], target: Callable, view: View):
        name = parent | 'root'
        return (
            self._layout_lens_bound(name) /
            target /
            __.modify(__.cat(view)) /
            self.with_sub
        )

    def _pane(self, ident: Ident):
        return self.state.vim_window // __.root.find_pane(ident)

    def _pane_uuid(self, ident: Ident):
        return self._pane(ident) / _.uuid

    def _default_pane_name(self):
        return 'make'

    def _pane_mod(self, ident: Ident, f: Callable[[Pane], Pane]):
        def g(path):
            return Task.now(Right(path.map_pane(f)))
        return _ident_ppm(ident) / g

    @property
    def _vim_pid(self):
        return self.vim.call('getpid').to_either('no pid')

    def _find_vim_pane(self, vim_pid):
        return self.server.pane_data.find(__.command_pid.contains(vim_pid))

    def _run_command_line(self, line, pane, options):
        return self.panes.run_command_line(pane, line)

    def _run_command(self, command, options):
        return self._run_command_ppm(command, options)

    def _run_in_shell(self, command: Command, shell: Shell, options: Map):
        ident = shell.target | self._default_pane_name
        opt = options + ('pane', ident) + ('shell', shell)
        cmd_runner = self._run_command(command, opt)
        shell_runner = TmuxRunShell(shell, Map())
        return cmd_runner.cons(shell_runner)


class Plugin(MyoComponent):

    Transitions = TmuxTransitions

    @lazy
    def native_server(self):
        socket = self.vim.pvar('tmux_socket') | None
        return libtmux.Server(socket_name=socket)

    @property
    def server(self):
        return Server(self.native_server)

    @property
    def layouts(self):
        return LayoutFacade(self.server)

    @property
    def panes(self):
        return PaneFacade(self.server)

    def new_state(self):
        return TmuxState(server=self.server)

    @lazy
    def watcher(self):
        interval = self.vim.pvar('tmux_watcher_interval') | 1.0
        return Watcher(self, interval=interval)

__all__ = ('Plugin',)



from amino import Path, __, Just

from ribosome.record import list_field, field, maybe_field, bool_field

from myo.record import Record, Named
from myo.ui.tmux.util import ident_field, Ident


class Command(Named):
    line = field(str)
    log_path = maybe_field(Path)
    parser = maybe_field(str)
    transient = bool_field()
    langs = list_field()


class VimCommand(Command):
    pass


class ShellCommand(Command):
    shell = ident_field()


class Shell(ShellCommand):
    target = ident_field()


class Commands(Record):
    commands = list_field()
    history = list_field()
    transient_cmd = maybe_field(Command)

    def __add__(self, command: Command):
        return self.append1.commands(command)

    def __getitem__(self, ident: Ident):
        return (
            self.commands.find(__.has_ident(ident))
            .to_either('no command with ident {}'.format(ident))
        )

    def __str__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ','.join(map(repr, self.commands))
        )

    def shell(self, ident):
        pred = lambda a: isinstance(a, Shell) and a.has_ident(ident)
        return self.commands.find(pred)

    def add_history(self, uuid):
        return self.append1.history(uuid)

    def transient_command(self, cmd):
        return self.set(transient_cmd=Just(cmd))

    def latest_command(self):
        def fetch(ident):
            return (
                self.transient_cmd.filter(__.has_ident(ident))
                .or_else(lambda: self[ident])
            )
        return self.history.head // fetch

__all__ = ('Commands', 'Command')



import libtmux

from ribosome.record import field, Record

from amino import List, __, _
from amino.lazy import lazy

from myo.ui.tmux.adapter import Adapter
from myo.ui.tmux.window import WindowAdapter
from myo.ui.tmux.util import parse_session_id


class Session(Record):
    id = field(str)


class SessionHandler:

    def __init__(self, server: libtmux.Server) -> None:
        self.server = server

    def create_pane(self, pane):
        pass


class SessionAdapter(Adapter):

    @lazy
    def id(self):
        return self.native.id

    @property
    def id_i(self):
        return parse_session_id(self.id)

    @lazy
    def windows(self):
        return List.wrap(self.native.windows) / WindowAdapter

    def window_by_id(self, id):
        return self.windows.find(__.id_i.contains(id))

    def find_pane_by_id(self, id: int):
        return self.windows.find_map(__.pane_by_id(id))

__all__ = ('Session',)



from subprocess import Popen, PIPE, STDOUT

import libtmux
from libtmux.exc import LibTmuxException

import amino.test
from amino.test.path import fixture_path
from amino import List, _
from amino.test import later
from amino.lazy import lazy

from myo.ui.tmux.server import Server


class Spec(amino.test.Spec):
    pass


class TmuxSpecBase(Spec):

    def setup(self):
        self.win_width = 100
        self.win_height = 40
        super().setup()

    @property
    def socket(self):
        return 'myo_spec'

    def _find_server(self):
        if self.native_server is None:
            try:
                self.native_server = libtmux.Server(socket_name=self.socket)
            except LibTmuxException:
                pass
        return self.native_server is not None

    def _find_session(self):
        try:
            return self.session is not None
        except LibTmuxException:
            pass

    @property
    def external_terminal(self):
        return True

    @property
    def _terminal_args(self):
        geom = '{}x{}'.format(self.win_width + 1, self.win_height + 1)
        return ['urxvt', '-geometry', geom, '-e']

    @property
    def _tmux_args(self):
        conf = fixture_path('conf', 'tmux.conf')
        return ['tmux', '-L', self.socket, '-f', str(conf)]

    @lazy
    def term(self):
        t = self._terminal_args if self.external_terminal else []
        args = t + self._tmux_args
        return Popen(args, stdout=PIPE, stderr=STDOUT)

    def _setup_server(self):
        self.native_server = None
        self.term
        self._wait_for(self._find_server)
        self._wait_for(self._find_session)

    def _teardown_server(self):
        self.term.kill()
        self.server.kill()

    @property
    def server(self):
        return Server(self.native_server)

    @property
    def session(self):
        return self.server.sessions[0]

    @property
    def sessions(self):
        return self.server.sessions

    @property
    def _panes(self):
        return self.sessions.head // _.windows.head / _.panes | List()

    @property
    def _output(self):
        return self._panes // _.capture

    def _output_contains(self, target):
        later(lambda: self._output.should.contain(target))

__all__ = ('Spec', 'TmuxSpecBase')



from typing import Tuple

from networkx import DiGraph

from amino.lazy import lazy
from amino import List, Empty, Just, Maybe, __, L, _

from ribosome.record import field, maybe_field

from myo.output.data import (PositionEntry, ErrorEntry, OutputEntry,
                             OutputEvent, OutputLine)
from myo.output.parser.base import EdgeData, SimpleParser


class FileEntry(PositionEntry):
    fun = field(str)
    expr = maybe_field(str)

    @property
    def output_lines(self):
        x = self.expr / L(OutputLine.create)(_, self)
        return super().output_lines + x.to_list


class ExprEntry(OutputEntry):
    pass

_file = EdgeData(
    r='  File "(?P<path>.+)", line (?P<line>\d+), in (?P<fun>\S+)',
    entry=FileEntry
)
_expr = EdgeData(r='    (.+)', entry=ExprEntry)
_error = EdgeData(r='(?P<error>\S+): (?P<msg>.+)', entry=ErrorEntry)


class Parser(SimpleParser):

    @lazy
    def graph(self):
        g = DiGraph()
        g.add_edge('start', 'file', data=_file)
        g.add_edge('file', 'expr', data=_expr)
        g.add_edge('expr', 'file', data=_file)
        g.add_edge('expr', 'error', data=_error)
        return g

    def event(self, entries: List[OutputEntry]):
        def folder(z: Tuple[List[OutputEntry], Maybe[FileEntry]], a):
            res, cur = z
            add, new = (
                (cur, Just(a))
                if isinstance(a, FileEntry) else
                ((cur / __.set(expr=Just(a.text))).or_else(Just(a)), Empty())
                if isinstance(a, ExprEntry) else
                (Just(a), Empty())
            )
            return res + add.to_list, new
        grouped, rest = entries.fold_left((List(), Empty()))(folder)
        complete = grouped + rest.to_list
        return Just(OutputEvent(head='traceback', entries=complete))

__all__ = ('Parser',)



from integration._support.base import MyoIntegrationSpec

from amino.test.path import fixture_path
from amino import Map, Just, List
from amino.test import temp_file
from amino.lazy import lazy

from ribosome.test.integration import main_looped
from ribosome.machine.scratch import Mapping

from myo.command import Command
from myo.plugins.core.message import ParseOutput


class ParseSpec(MyoIntegrationSpec):

    @main_looped
    def native(self):
        cmd = Command(name='a', line='a', parser=Just('compiler'))
        fix = fixture_path('parse', 'err1')
        output = List.wrap(fix.read_text().splitlines())
        errorformat = List(
            '%AIn %f:%l',
            '%C  %.%#',
            '%+Z%.%#Error: %.%#'
        ).mk_string(',')
        self.vim.buffer.set_option('errorformat', errorformat)
        msg = ParseOutput(cmd, output, fix, Map())
        self.plugin.myo_start()
        self.root.send(msg)
        self._await()
        qf = self.vim.call('getqflist') | []
        qf.should_not.be.empty

    @property
    def _python_func_name(self):
        return 'func_name'

    @lazy
    def _python_code(self):
        return List(
            'import functools',
            'def {}(a, b=1):'.format(self._python_func_name),
            '    return 1'
        )

    @lazy
    def _python_file(self):
        path = temp_file('core', 'parse', 'python', 'code.py')
        path.write_text(self._python_code.join_lines)
        return path

    @property
    def _mk_trace(self):
        r = lambda: List.random_string(4)
        tmpl = '  File "{}", line 2, in {}'
        frame = lambda i: List(
            tmpl.format(self._python_file, self._python_func_name),
            '    {}()'.format(r())
        )
        t = List.range(3) // frame
        err = 'AttributeError: butt'
        return t.cat(err)

    def python(self):
        trace1 = self._mk_trace
        trace2 = self._mk_trace
        output = trace1 + trace2
        cmd = Command(name='a', line='a', langs=List('python'))
        msg = ParseOutput(cmd, output, None, Map())
        self.plugin.myo_start()
        self.root.send_sync(msg)
        output_machine = self.root.sub[-1]
        self.vim.buffer.content.should.contain(output[-1])
        self.vim.buffer.option('modifiable').should.contain(False)
        self.vim.window.set_cursor(4)
        self.root.send_sync(Mapping(output_machine.uuid, '%cr%'))
        self.vim.windows.should.have.length_of(2)
        self.vim.window.buffer.name.should.equal(str(self._python_file))
        self.vim.window.cursor.should.equal(List(2, 0))

__all__ = ('ParseSpec',)



from integration._support.tmux import TmuxCmdSpec

from psutil import Process
from amino.test import later


class DispatchSpec(TmuxCmdSpec):

    def simple(self):
        target = 'cmd test'
        self.json_cmd('MyoShellCommand test', line="echo '{}'".format(target))
        self.json_cmd('MyoTmuxOpenPane make')
        self.json_cmd('MyoRun test', pane='make')
        self._output_contains(target)

    def _shell(self, create):
        s = 'cmd test {}'
        i = 5342
        target = s.format(i)
        line = 'print(\'{}\'.format(\'{}\'))'.format(s, i)
        self.json_cmd('MyoTmuxCreatePane py', parent='main')
        self.json_cmd('MyoShell py', line='python', target='py')
        create(line)
        self._output_contains(target)

    def run_in_shell(self):
        def create(line):
            self.json_cmd('MyoRun py')
            self.json_cmd('MyoRunInShell py', line=line)
        self._shell(create)

    def kill_shell_command(self):
        def create(line):
            self.json_cmd('MyoShellCommand test', line=line, shell='py')
            self.json_cmd('MyoRun test')
        def pid():
            expr = '''data.pane("py") // (lambda a: a.pid) | -1'''
            return self.vim.call('MyoTmuxEval', expr) | -2
        self._shell(create)
        later(lambda: pid().should.be.greater_than(0))
        Process(pid()).kill()
        later(lambda: pid().should.equal(-1))

__all__ = ('DispatchSpec',)



from amino import _, Path, List
from amino.lazy import lazy

from ribosome.record import field, list_field, dfield, any_field

from myo.record import Record
from myo.util import parse_int


class OutputEntry(Record):
    text = field(str)

    @property
    def output_lines(self):
        return List(OutputLine.create(self.text, self))


class OutputLine(Record):
    text = field(str)
    target = any_field()

    @staticmethod
    def create(text, target):
        return OutputLine(text=text, target=target)


class ErrorEntry(OutputEntry):
    error = field(str)
    msg = field(str)


def _parse_line(data):
    return data if isinstance(data, int) else parse_int(data) | 0


class PositionEntry(OutputEntry):
    path = field(Path, factory=Path)
    line = field(int, factory=_parse_line)
    col = dfield(0)


class OutputEvent(Record):
    head = field(str)
    entries = list_field(OutputEntry)

    @property
    def lines(self):
        return (
            (self.entries // _.output_lines)
            .cons(OutputLine.create(self.head, self))
        )


class ParseResult(Record):
    head = field(str)
    events = list_field(OutputEvent)

    @lazy
    def lines(self):
        return (
            (self.events // _.lines)
            .cons(OutputLine.create(self.head, self))
        )

    @property
    def display_lines(self):
        return self.lines / _.text

    def target_for_line(self, line):
        return self.lines.lift(line) / _.target

__all__ = ('OutputEntry', 'OutputEvent', 'ParseResult')



from pathlib import Path

import neovim

from amino import List, L, _, I

from ribosome import command, NvimStatePlugin, msg_command, json_msg_command
from ribosome.machine.scratch import Mapping
from ribosome.request import msg_function

from myo.plugins.core.main import StageI
from myo.main import Myo
from myo.logging import Logging
from myo.plugins.command.message import (AddVimCommand, Run, AddShellCommand,
                                         AddShell, ShellRun, RunTest,
                                         RunVimTest)
from myo.plugins.tmux.message import (TmuxCreatePane, TmuxCreateSession,
                                      TmuxCreateLayout, TmuxSpawnSession,
                                      TmuxInfo, TmuxClosePane, TmuxPack)
from myo.plugins.tmux import TmuxOpenPane
from myo.plugins.core.message import Parse


class MyoNvimPlugin(NvimStatePlugin, Logging):

    def __init__(self, vim: neovim.Nvim) -> None:
        super().__init__(vim)
        self.myo = None  # type: Myo
        self._post_initialized = False

    @property
    def name(self):
        return 'myo'

    @property
    def state(self):
        return self.myo

    @command()
    def myo_reload(self):
        self.myo_quit()
        self.myo_start()

    @command()
    def myo_quit(self):
        if self.myo is not None:
            self.myo.stop()
            self.vim.clean()
            self.myo = None

    @command(sync=True)
    def myo_start(self):
        config_path = self.vim.ppath('config_path')\
            .get_or_else(Path('/dev/null'))
        plugins = self.vim.pl('plugins') | List()
        self.myo = Myo(self.vim.proxy, Path(config_path), plugins)
        self.myo.start()
        self.myo.wait_for_running()
        self.myo.send(StageI())

    @command()
    def myo_post_startup(self):
        self._post_initialized = True
        if self.myo is not None:
            pass
        else:
            self.log.error('myo startup failed')

    @command()
    def myo_plug(self, plug_name, cmd_name, *args):
        self.myo.plug_command(plug_name, cmd_name, args)

    @json_msg_command(AddVimCommand)
    def myo_vim_command(self):
        pass

    @json_msg_command(AddShellCommand)
    def myo_shell_command(self):
        pass

    @json_msg_command(AddShell)
    def myo_shell(self):
        pass

    @json_msg_command(TmuxCreateSession)
    def myo_tmux_create_session(self):
        pass

    @msg_command(TmuxSpawnSession)
    def myo_tmux_spawn_session(self):
        pass

    @json_msg_command(TmuxCreateLayout)
    def myo_tmux_create_layout(self):
        pass

    @json_msg_command(TmuxCreatePane)
    def myo_tmux_create_pane(self):
        pass

    @json_msg_command(TmuxOpenPane)
    def myo_tmux_open_pane(self):
        pass

    @msg_command(TmuxClosePane)
    def myo_tmux_close_pane(self):
        pass

    @msg_command(TmuxPack)
    def myo_tmux_pack(self):
        pass

    @json_msg_command(Run)
    def myo_run(self):
        pass

    @json_msg_command(ShellRun)
    def myo_run_in_shell(self):
        pass

    @msg_command(TmuxInfo)
    def myo_tmux_show(self):
        pass

    @json_msg_command(Parse)
    def myo_parse(self):
        pass

    @json_msg_command(RunTest)
    def myo_test(self):
        pass

    @json_msg_command(RunVimTest)
    def myo_vim_test(self):
        pass

    def _eval(self, args, go):
        result = List.wrap(args).head.to_either('expression needed') // go
        return result.cata(self.log.error, I)

    @neovim.function('MyoEval', sync=True)
    def myo_eval(self, args):
        return self._eval(args, self.myo.eval_expr)

    @neovim.function('MyoTmuxEval', sync=True)
    def myo_tmux_eval(self, args):
        def mod(data, plugins):
            d = data.sub_states.get('tmux') | data
            p = plugins.get('tmux') | plugins
            return d, p
        return self._eval(args, L(self.myo.eval_expr)(_, mod))

    @neovim.autocmd('VimLeave', sync=True)
    def vim_leave(self):
        self.myo_quit()

    @msg_function(Mapping)
    def myo_mapping(self):
        pass

__all__ = ('MyoNvimPlugin',)



from amino import List, _
from amino.test import later

from integration._support.tmux import TmuxIntegrationSpec, DefaultLayoutSpec


class CutSizeSpec(TmuxIntegrationSpec):

    def cut_size(self):
        def check():
            panes = self.sessions.head // _.windows.head / _.panes | List()
            panes.should.have.length_of(2)
        self.json_cmd('MyoTmuxCreatePane pan', parent='root', min_size=0.5,
                      weight=0.1)
        self.json_cmd('MyoTmuxOpenPane pan')
        later(check)


class DistributeSizeSpec(TmuxIntegrationSpec):

    def distribute(self):
        h1 = 7
        def check():
            panes = self.sessions.head // _.windows.head / _.panes | List()
            sizes = panes[1:] / _.size / _[1]
            sizes.head.should.contain(h1)
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self.json_cmd('MyoTmuxCreatePane pan1', parent='test', min_size=5,
                      max_size=h1, weight=1)
        self._wait(.1)
        self.json_cmd('MyoTmuxCreatePane pan2', parent='test', min_size=10,
                      max_size=40, weight=9)
        self._wait(.1)
        self.json_cmd('MyoTmuxOpenPane pan1')
        self._wait(.1)
        self.json_cmd('MyoTmuxOpenPane pan2')
        later(check)


class DefaultLayoutDistributeSizeSpec(DefaultLayoutSpec):

    def _set_vars(self):
        super()._set_vars()
        self.vim_width = 10
        self.vim.set_pvar('tmux_vim_width', self.vim_width)

    def distribute(self):
        def check():
            widths = self.sessions.head // _.windows // _.panes / _.size[0]
            target = List(self.vim_width, self.win_width - self.vim_width)
            widths.should.equal(target)
        self.json_cmd('MyoTmuxOpenPane make')
        later(check)

__all__ = ('CutSizeSpec', 'DistributeSizeSpec',
           'DefaultLayoutDistributeSizeSpec')



from integration._support.tmux import DefaultLayoutSpec, TmuxIntegrationSpec

from amino.test import later


class ClosePaneSpec(DefaultLayoutSpec):

    def _check(self, length: int):
        return later(lambda: self._panes.should.have.length_of(length))

    def close(self):
        self.json_cmd('MyoTmuxOpenPane make')
        self._check(2)
        self.vim.cmd('MyoTmuxClosePane make')
        self._check(1)
        self.vim.cmd('MyoTmuxClosePane make')
        self._wait(0.1)
        self._check(1)
        self.json_cmd('MyoTmuxOpenPane make')
        self._check(2)
        self.vim.cmd('MyoTmuxClosePane make')
        self._check(1)

    def auto_close(self):
        self.json_cmd('MyoTmuxOpenPane make')
        self._check(2)
        self.vim.doautocmd('VimLeave')
        self._check(1)


class ShowSpec(TmuxIntegrationSpec):

    def output(self):
        def check():
            self._log_out.should_not.be.empty
        self.vim.cmd('MyoTmuxShow')
        later(check)

__all__ = ('ClosePaneSpec', 'ShowSpec')



from operator import add

from ribosome.machine import message, may_handle, Machine, handle
from ribosome.nvim import ScratchBuffer, NvimFacade
from ribosome.machine.scratch import ScratchMachine

from amino.task import Task
from amino import Map, _, L, Left, __

from myo.output.data import ParseResult
from myo.state import MyoTransitions
from myo.logging import Logging
from myo.output.parser.python import FileEntry

OutputInit = message('OutputInit')
Jump = message('Jump')


class OutputMachineTransitions(MyoTransitions):

    @property
    def scratch(self):
        return self.machine.scratch

    @property
    def buffer(self):
        return self.scratch.buffer

    @property
    def result(self):
        return self.machine.result

    @may_handle(OutputInit)
    def output_init(self):
        return (
            Task.call(self.buffer.set_content, self.result.display_lines) //
            (lambda a: Task.call(self.buffer.set_modifiable, False))
        )

    @handle(Jump)
    def jump(self):
        target = (
            self.vim.window.line /
            (_ - 1) //
            self.result.target_for_line
        ).filter_type(FileEntry).to_either('not a file entry')
        open_file = target / L(self._open_file)(_.path)
        set_line = (target / _.line /
                    (lambda a: Task(lambda: self.vim.window.set_cursor(a))))
        return (open_file & set_line).map2(add)

    def _open_file(self, path):
        def split():
            # TODO
            pass
        if not path.is_file():
            return Task.now(Left('not a file: {}'.format(path)))
        else:
            win = (
                self.vim.windows.find(__.buffer.modifiable.contains(True))
                .or_else(split)
                .task('could not get a window') /
                __.focus()
            )
            edit = Task.call(self.vim.edit, path) / __.run_async()
            return win + edit


class OutputMachine(ScratchMachine, Logging):
    Transitions = OutputMachineTransitions

    def __init__(self, vim: NvimFacade, scratch: ScratchBuffer,
                 result: ParseResult, parent: Machine) -> None:
        super().__init__(vim, scratch, parent=parent, title='output')
        self.result = result

    @property
    def prefix(self):
        return 'Myo'

    @property
    def mappings(self):
        special = {
            '%cr%': Jump,
        }
        return Map(**special)

__all__ = ('OutputMachine',)



from ribosome.machine import message, json_message

Run = json_message('Run', 'command')
ShellRun = json_message('ShellRun', 'shell')
RunTest = json_message('RunTest')
RunVimTest = json_message('RunVimTest')
Dispatch = json_message('Dispatch', 'command')
AddCommand = json_message('AddCommand', 'name')
AddShellCommand = json_message('AddShellCommand', 'name')
AddShell = json_message('AddShell', 'name')
AddVimCommand = json_message('AddVimCommand', 'name')
SetShellTarget = message('SetShellTarget', 'shell', 'target')
CommandExecuted = message('CommandExecuted', 'command')

__all__ = ('Run', 'ShellRun', 'Dispatch', 'AddCommand', 'AddShellCommand',
           'AddShell', 'AddVimCommand', 'SetShellTarget', 'CommandExecuted',
           'RunTest', 'RunVimTest')



import libtmux

from myo.logging import Logging
from myo.ui.tmux.session import SessionAdapter
from myo.ui.tmux.pane import PaneData

from amino import _, List, __, Map
from amino.task import task
from amino.lazy import lazy


class Server(Logging):

    def __init__(self, native: libtmux.Server) -> None:
        self.native = native

    @lazy
    def sessions(self):
        return List.wrap(self.native.sessions) / SessionAdapter

    def session_by_id(self, id: int):
        return self.sessions.find(__.id_i.contains(id))

    @lazy
    def windows(self):
        return self.sessions // _.windows

    def window(self, id):
        return self.windows.find(__.id_i.contains(id))

    @lazy
    def panes(self):
        return self.windows // _.panes

    def pane(self, id):
        return self.panes.find(__.id_i.contains(id))

    @lazy
    def pane_data(self):
        return List.wrap(self.native._list_panes()) / Map / PaneData

    @lazy
    def pane_ids(self):
        return self.pane_data // _.id_i.to_list

    def kill(self):
        return self.native.kill_server()

    @task
    def cmd(self, *a, **kw):
        return self.native.cmd(*a, **kw)

__all__ = ('Server',)



from typing import Tuple
from functools import singledispatch  # type: ignore
from operator import ne, sub

from amino.task import Task
from amino.anon import L
from amino import _, __, List, Left, F, Boolean, Right, Empty, Either, Maybe
from amino.lazy import lazy

from myo.logging import Logging
from myo.ui.tmux.server import Server
from myo.ui.tmux.window import Window
from myo.ui.tmux.layout import Layout
from myo.ui.tmux.pane import Pane, PaneAdapter
from myo.command import ShellCommand
from myo.ui.tmux.pane_path import PanePath


class LayoutFacade(Logging):

    def __init__(self, server: Server) -> None:
        self.server = server
        self.panes = PaneFacade(server)

    def layout_open(self, layout):
        return (layout.layouts.exists(self.layout_open) or
                layout.panes.exists(self.panes.is_open))

    def view_open(self, view):
        f = (self.layout_open if isinstance(view, Layout) else
             self.panes.is_open)
        return f(view)

    def open_views(self, layout):
        return layout.views.filter(self.view_open)

    def open_pane(self, path: PanePath) -> Task[Either[str, PanePath]]:
        return (
            self._pane_opener(path).eff(Either) //
            L(self._split_pane)(_, path.pane).map(Right) /
            path.setter.pane
        ).value

    def _pane_opener(self, path: PanePath) -> Task[Either[str, Layout]]:
        pane = path.pane
        def find_layout():
            return (path.layouts.reversed.find(self.layout_open)
                    .to_either('no open layout when trying to open pane'))
        return Task.now(
            self.panes.is_open(pane).no.flat_either_call(
                'pane {} already open'.format(pane),
                find_layout,
            )
        )

    def _split_pane(self, layout, pane) -> Task[Pane]:
        return self._ref_pane(layout) / (
            lambda a:
            self.panes.find(a) /
            __.split(layout.horizontal) /
            PaneAdapter /
            pane.open |
            pane
        )

    def _ref_pane(self, layout) -> Task[Pane]:
        ''' A pane in the layout to be used to open a pane.
        Uses the first open pane available.
        Failure if no panes are open.
        '''
        def go(l):
            return (self._opened_panes(l.panes).head
                    .or_else(l.layouts.find_map(go)))
        return (Task.call(go, layout) //
                L(Task.from_maybe)(_, "no ref pane for {}".format(layout)))

    def open_pane_path(self, path: PanePath) -> Task[Either[str, PanePath]]:
        ''' legacy version
        '''
        p = path.pane
        return (Task.now(Left('pane {} already open'.format(p)))
                if self.panes.is_open(p)
                else self._open_pane_path(path))

    def _open_in_layouts_path(self, pane, layout, outer):
        if self.layout_open(layout):
            return self._open_in_layout(pane, layout) / (_ + (outer,))
        else:
            return (
                outer.detach_last.map2(F(self._open_in_layouts_path, pane)) |
                Task.failed('cannot open {} without open layout'.format(pane))
            ).map3(lambda p, l, o: (p, layout, o.cat(l)))

    def _open_in_layout(self, pane, layout) -> Task[Tuple[Layout, Pane]]:
        return self._split_pane(layout, pane) & (Task.now(layout))

    def close_pane_path(self, path: PanePath):
        new_path = path.map_pane(__.set(id=Empty()))
        return self.panes.close(path.pane) / __.replace(new_path)

    def _opened_panes(self, panes):
        return panes.filter(self.panes.is_open)

    def pack_path(self, path: PanePath):
        return self.pack_window(path.window) / __.replace(path)

    def pack_window(self, window: Window) -> Task[Either[str, Window]]:
        t = (
            Task.call(window.id.flat_map, self.server.window) //
            L(Task.from_maybe)(_, 'window not found: {}'.format(window)) /
            _.size
        )
        return (
            t.flat_map2(L(self._pack_layout)(window.root, _, _)) /
            Right /
            __.replace(window)
        )

    def _pack_layout(self, l, w, h):
        horizontal = l.horizontal
        total = w if horizontal else h
        @singledispatch
        def recurse(v, size):
            pass
        @recurse.register(Pane)
        def rec_pane(v, size):
            return self._apply_size(v, size, horizontal)
        @recurse.register(Layout)
        def rec_layout(v, size):
            sub_size = (size, h) if horizontal else (w, size)
            return (
                self._pack_layout(v, *sub_size)
                .flat_replace(
                    self._ref_pane(v) //
                    L(self._apply_size)(_, size, horizontal)
                )
            )
        views = self.open_views(l)
        count = views.length
        if count > 0:
            # each pane spacer takes up one cell
            m = self._measure_layout(views, total - count + 1)
            return views.zip(m).map2(recurse).sequence(Task)
        else:
            return Task.now(List())

    def _measure_layout(self, views, total):
        calc = lambda s: s if s > 1 else s * total
        min_s = self.actual_min_sizes(views) / calc
        max_s = self.actual_max_sizes(views) / calc
        weights = self.weights(views)
        return (self._cut_sizes(min_s, weights, total) if sum(min_s) > total
                else self._distribute_sizes(min_s, max_s, weights, total))

    def _cut_sizes(self, min_s, weights, total):
        surplus = sum(min_s) - total
        dist = weights / (1 - _) / (surplus * _)
        return (min_s & dist).map2(sub)

    def _distribute_sizes(self, min_s, max_s, weights, total):
        count = max_s.length
        surplus = total - sum(min_s)
        sizes1 = min_s.zip(max_s, weights).map3(
            lambda l, h, w: min(l + w * surplus, h))
        rest = total - sum(sizes1)
        unsat = (max_s & sizes1).map2(ne) / Boolean
        unsat_left = count - sum(unsat / _.value) > 0
        def trim_weights():
            w1 = (unsat & weights).map2(lambda s, w: s.maybe(w))
            return self._normalize_weights(w1)
        def dist_rest():
            rest_w = trim_weights() if unsat_left else weights
            return (sizes1 & rest_w).map2(lambda a, w: a + w * rest)
        return sizes1 if rest <= 0 else dist_rest()

    def _apply_size(self, pane, size, horizontal):
        return (
            Task.call(self.panes.find, pane) //
            L(Task.from_maybe)(_, 'pane not found: {}'.format(pane)) //
            __.resize(size, horizontal)
        )

    def actual_min_sizes(self, views):
        return views / (lambda a: a.fixed_size.or_else(a.min_size) | 0)

    def actual_max_sizes(self, views):
        return views / (lambda a: a.fixed_size.or_else(a.max_size) | 999)

    def weights(self, views):
        return self._normalize_weights(views / _.effective_weight)

    def _normalize_weights(self, weights):
        amended = self._amend_weights(weights)
        total = sum(amended)
        total1 = 1 if total == 0 else total
        return amended / (_ / total1)

    def _amend_weights(self, weights):
        total = sum(weights.join)
        total1 = 1 if total == 0 else total
        empties = weights.filter(_.is_empty).length
        empties1 = 1 if empties == 0 else empties
        empty = total1 / empties1
        return weights / (_ | empty)


class PaneFacade(Logging):

    def __init__(self, server: Server) -> None:
        self.server = server

    @lazy
    def panes(self):
        return self.server.panes

    @lazy
    def pane_data(self):
        return self.server.pane_data

    @lazy
    def pane_ids(self):
        return self.pane_data // _.id_i.to_list

    def is_open(self, pane):
        return pane.id.exists(self.pane_ids.contains)

    def find_by_id(self, id):
        return self.panes.find(__.id_i.contains(id))

    def find_by_name(self, name):
        return self.panes.find(_.name == name)

    def select_pane(self, session_id, window_id, pane_id):
        return (
            self.server.session_by_id(session_id) //
            __.window_by_id(window_id) //
            __.pane_by_id(pane_id)
        )

    def find(self, pane: Pane) -> Maybe[PaneAdapter]:
        def find_by_id(pane_id):
            return (
                (pane.session_id & pane.window_id)
                .flat_map2(L(self.select_pane)(_, _, pane_id))
                .or_else(L(self.find_by_id)(pane_id))
            )
        return pane.id // find_by_id

    def _find_task(self, pane: Pane) -> Task[PaneAdapter]:
        return (
            Task.call(self.find, pane) //
            L(Task.from_maybe)(_, 'pane not found')
        )

    def run_command(self, pane: Pane, command: ShellCommand):
        return self.run_command_line(pane, command.line)

    def run_command_line(self, pane: Pane, line: str):
        return (
            self._find_task(pane) /
            __.send_keys(line, suppress_history=False)
        )

    def command_pid(self, pane: Pane):
        return self._find_task(pane) / _.command_pid

    def close(self, pane: Pane):
        err = 'cannot close pane {}: not found'.format(pane)
        return (Task.call(self.find, pane) //
                __.cata(__.kill.map(Right), lambda: Task.now(Left(err))))

    def close_id(self, id: int):
        return self.server.cmd('kill-pane', '-t', '%{}'.format(id))

    def ensure_not_running(self, pane: Pane):
        return (
            self._find_task(pane) /
            __.not_running.either('command already running', True)
        )

    def pipe(self, pane: Pane, base: str):
        return (
            self.find(pane)
            .task('capture: pane not found') /
            __.pipe(base) /
            pane.setter.log_path
        )

__all__ = ('LayoutFacade', 'PaneFacade')
